from __future__ import absolute_import, division, print_function

import uuid

import numpy as np
from glue.viewers.image.state import ImageLayerState, ImageSubsetLayerState
from glue.viewers.scatter.state import ScatterLayerState

STATE_CLASS = {}
STATE_CLASS['ImageLayerArtist'] = ImageLayerState
STATE_CLASS['ScatterLayerArtist'] = ScatterLayerState
STATE_CLASS['SubsetImageLayerArtist'] = ImageSubsetLayerState


class DS9Compat(object):

    @classmethod
    def __setgluestate__(cls, rec, context):
        result = cls()
        for k, v in rec.items():
            setattr(result, k, v)
        return result


def update_image_viewer_state(rec, context):
    """
    Given viewer session information, make sure the session information is
    compatible with the current version of the viewers, and if not, update
    the session information in-place.
    """

    if '_protocol' not in rec:

        # Note that files saved with protocol < 1 have bin settings saved per
        # layer but they were always restricted to be the same, so we can just
        # use the settings from the first layer

        rec['state'] = {}
        rec['state']['values'] = {}

        # TODO: could generalize this into a mapping
        properties = rec.pop('properties')
        viewer_state = rec['state']['values']
        viewer_state['color_mode'] = 'st__Colormaps'
        viewer_state['reference_data'] = properties['data']

        data = context.object(properties['data'])

        # TODO: add an id method to unserializer

        x_index = properties['slice'].index('x')
        y_index = properties['slice'].index('y')

        viewer_state['x_att_world'] = str(uuid.uuid4())
        context.register_object(viewer_state['x_att_world'], data.world_component_ids[x_index])

        viewer_state['y_att_world'] = str(uuid.uuid4())
        context.register_object(viewer_state['y_att_world'], data.world_component_ids[y_index])

        viewer_state['x_att'] = str(uuid.uuid4())
        context.register_object(viewer_state['x_att'], data.pixel_component_ids[x_index])

        viewer_state['y_att'] = str(uuid.uuid4())
        context.register_object(viewer_state['y_att'], data.pixel_component_ids[y_index])

        viewer_state['x_min'] = -0.5
        viewer_state['x_max'] = data.shape[1] - 0.5
        viewer_state['y_min'] = -0.5
        viewer_state['y_max'] = data.shape[0] - 0.5

        viewer_state['aspect'] = 'st__equal'

        # Slicing with cubes
        viewer_state['slices'] = [s if np.isreal(s) else 0 for s in properties['slice']]

        # RGB mode
        for layer in rec['layers'][:]:
            if layer['_type'].split('.')[-1] == 'RGBImageLayerArtist':
                for icolor, color in enumerate('rgb'):
                    new_layer = {}
                    new_layer['_type'] = 'glue.viewers.image.layer_artist.ImageLayerArtist'
                    new_layer['layer'] = layer['layer']
                    new_layer['attribute'] = layer[color]
                    new_layer['norm'] = layer[color + 'norm']
                    new_layer['zorder'] = layer['zorder']
                    new_layer['visible'] = layer['color_visible'][icolor]
                    new_layer['color'] = color
                    rec['layers'].append(new_layer)
                rec['layers'].remove(layer)
                viewer_state['color_mode'] = 'st__One color per layer'

        layer_states = []

        for layer in rec['layers']:

            state_id = str(uuid.uuid4())
            state_cls = STATE_CLASS[layer['_type'].split('.')[-1]]
            state = state_cls(layer=context.object(layer.pop('layer')))
            for prop in ('visible', 'zorder'):
                value = layer.pop(prop)
                value = context.object(value)
                setattr(state, prop, value)
            if 'attribute' in layer:
                state.attribute = context.object(layer['attribute'])
            else:
                state.attribute = context.object(properties['attribute'])
            if 'norm' in layer:
                norm = context.object(layer['norm'])
                state.bias = norm.bias
                state.contrast = norm.contrast
                state.stretch = norm.stretch
                if norm.clip_hi is not None:
                    state.percentile = norm.clip_hi
                else:
                    if norm.vmax is not None:
                        state.v_min = norm.vmin
                        state.v_max = norm.vmax
                        state.percentile = 'Custom'
            if 'color' in layer:
                state.global_sync = False
                state.color = layer['color']
            context.register_object(state_id, state)
            layer['state'] = state_id
            layer_states.append(state)

        list_id = str(uuid.uuid4())
        context.register_object(list_id, layer_states)
        rec['state']['values']['layers'] = list_id



from __future__ import absolute_import, division, print_function

import os
import numpy as np
from matplotlib import cm

from qtpy import QtWidgets, QtGui
from glue.core.util import colorize_subsets, facet_subsets
from glue.utils.qt import load_ui
from glue.utils.qt.widget_properties import (ButtonProperty, FloatLineProperty,
                                             ValueProperty)
from glue.utils.array import pretty_number
from glue.utils import Pointer, nanmin, nanmax
from glue.utils.qt import cmap2pixmap

# We do the following import to register the custom Qt Widget there
from glue.dialogs.common.qt import component_selector  # noqa

__all__ = ['SubsetFacet']


class SubsetFacet(QtWidgets.QDialog):

    log = ButtonProperty('ui.checkbox_log')
    vmin = FloatLineProperty('ui.value_min')
    vmax = FloatLineProperty('ui.value_max')
    steps = ValueProperty('ui.value_n_subsets')
    data = Pointer('ui.component_selector.data')
    component = Pointer('ui.component_selector.component')

    def __init__(self, collect, default=None, parent=None):
        """Create a new dialog for subset faceting

        :param collect: The :class:`~glue.core.data_collection.DataCollection` to use
        :param default: The default dataset in the collection (optional)
        """

        super(SubsetFacet, self).__init__(parent=parent)

        self.ui = load_ui('subset_facet.ui', self,
                          directory=os.path.dirname(__file__))
        self.ui.setWindowTitle("Subset Facet")
        self._collect = collect

        self.ui.component_selector.setup(self._collect)
        if default is not None:
            self.ui.component_selector.data = default

        val = QtGui.QDoubleValidator(-1e100, 1e100, 4, None)
        self.ui.component_selector.component_changed.connect(self._set_limits)

        combo = self.ui.color_scale
        for cmap in [cm.cool, cm.RdYlBu, cm.RdYlGn, cm.RdBu, cm.Purples]:
            combo.addItem(QtGui.QIcon(cmap2pixmap(cmap)), cmap.name, cmap)

    def _set_limits(self):
        data = self.ui.component_selector.data
        cid = self.ui.component_selector.component

        vals = data[cid]

        wmin = self.ui.value_min
        wmax = self.ui.value_max

        wmin.setText(pretty_number(nanmin(vals)))
        wmax.setText(pretty_number(nanmax(vals)))

    @property
    def cmap(self):
        combo = self.ui.color_scale
        index = combo.currentIndex()
        return combo.itemData(index)

    def _apply(self):
        try:
            lo, hi = self.vmin, self.vmax
        except ValueError:
            return  # limits not set. Abort
        if not np.isfinite(lo) or not np.isfinite(hi):
            return

        subsets = facet_subsets(self._collect, self.component, lo=lo, hi=hi,
                                steps=self.steps, log=self.log)
        colorize_subsets(subsets, self.cmap)

    @classmethod
    def facet(cls, collect, default=None, parent=None):
        """Class method to create facted subsets
        The arguments are the same as __init__
        """
        self = cls(collect, parent=parent, default=default)
        value = self.exec_()

        if value == QtWidgets.QDialog.Accepted:
            self._apply()



from __future__ import absolute_import, division, print_function

from glue.core.subset import SliceSubsetState

__all__ = ['PixelSubsetState']


class PixelSubsetState(SliceSubsetState):
    def copy(self):
        return PixelSubsetState(self.reference_data, self.slices)



from __future__ import absolute_import, division, print_function

from qtpy import compat
from glue import config
from glue.utils.qt import messagebox_on_error
from glue.io.subset_mask import SubsetMaskImporter, SubsetMaskExporter

__all__ = ['QtSubsetMaskImporter', 'QtSubsetMaskExporter']


def _make_filters_dict(registry):

    filters = {}
    for e in registry:
        if e.extension == '':
            fltr = "{0} (*)".format(e.label)
        else:
            fltr = "{0} ({1})".format(e.label, ' '.join('*.' + ext for ext in e.extension))
        filters[fltr] = e.function

    return filters


class QtSubsetMaskImporter(SubsetMaskImporter):

    def get_filename_and_reader(self):

        subset_mask_importers = _make_filters_dict(config.subset_mask_importer)

        filters = ';;'.join(sorted(subset_mask_importers))

        filename, fltr = compat.getopenfilename(caption="Choose a subset mask file",
                                                filters=filters)

        filename = str(filename)

        if filename:
            return filename, subset_mask_importers[fltr]
        else:
            return None, None

    @messagebox_on_error('An error occurred when importing the subset mask file:', sep=' ')
    def run(self, data_or_subset, data_collection):
        super(QtSubsetMaskImporter, self).run(data_or_subset, data_collection)


class QtSubsetMaskExporter(SubsetMaskExporter):

    def get_filename_and_writer(self):

        subset_mask_exporters = _make_filters_dict(config.subset_mask_exporter)

        filters = ';;'.join(sorted(subset_mask_exporters))

        filename, fltr = compat.getsavefilename(caption="Choose a subset mask filename",
                                                filters=filters)

        filename = str(filename)

        if filename:
            return filename, subset_mask_exporters[fltr]
        else:
            return None, None

    @messagebox_on_error('An error occurred when exporting the subset mask:', sep=' ')
    def run(self, data_or_subset):
        super(QtSubsetMaskExporter, self).run(data_or_subset)



from __future__ import absolute_import, division, print_function

import pytest
import numpy as np
from glue.core import Data
from glue.core.coordinates import WCSCoordinates
from astropy.io import fits
from astropy.wcs import WCS

from ..gridded_fits import fits_writer

BITPIX = {}
BITPIX[np.int16] = 16
BITPIX[np.int32] = 32
BITPIX[np.int64] = 64
BITPIX[np.float32] = -32
BITPIX[np.float64] = -64


@pytest.mark.parametrize('dtype', BITPIX.keys())
def test_fits_writer_data(tmpdir, dtype):

    dtype = np.dtype(dtype)

    filename = tmpdir.join('test1.fits').strpath

    data = Data(x=np.arange(6).reshape(2, 3).astype(dtype),
                y=(np.arange(6) * 2).reshape(2, 3).astype(dtype))

    fits_writer(filename, data)

    with fits.open(filename) as hdulist:
        assert len(hdulist) == 2
        np.testing.assert_equal(hdulist['x'].data, data['x'])
        np.testing.assert_equal(hdulist['y'].data, data['y'])
        # Note: the following tolerates endian-ness change
        assert hdulist['x'].data.dtype in (dtype, dtype.newbyteorder())
        assert hdulist['y'].data.dtype in (dtype, dtype.newbyteorder())

    # Only write out some components

    filename = tmpdir.join('test2.fits').strpath

    fits_writer(filename, data, components=[data.id['x']])

    with fits.open(filename) as hdulist:
        assert len(hdulist) == 1
        np.testing.assert_equal(hdulist['x'].data, data['x'])


def test_component_unit_header(tmpdir):
    from astropy import units as u
    filename = tmpdir.join('test3.fits').strpath

    data = Data(x=np.arange(6).reshape(2, 3),
                y=(np.arange(6) * 2).reshape(2, 3),
                z=(np.arange(6) * 2).reshape(2, 3))

    wcs = WCS()
    data.coords = WCSCoordinates(wcs=wcs)

    unit1 = data.get_component("x").units = u.m / u.s
    unit2 = data.get_component("y").units = u.Jy
    unit3 = data.get_component("z").units = ""

    fits_writer(filename, data)

    with fits.open(filename) as hdulist:
        assert len(hdulist) == 3
        bunit = hdulist['x'].header.get('BUNIT')
        assert u.Unit(bunit) == unit1

        bunit = hdulist['y'].header.get('BUNIT')
        assert u.Unit(bunit) == unit2

        bunit = hdulist['z'].header.get('BUNIT')
        assert bunit == unit3


@pytest.mark.parametrize('dtype', BITPIX.keys())
def test_fits_writer_subset(tmpdir, dtype):

    filename = tmpdir.join('test').strpath

    data = Data(x=np.arange(6).reshape(2, 3).astype(dtype),
                y=(np.arange(6) * 2).reshape(2, 3).astype(dtype))

    subset = data.new_subset()
    subset.subset_state = data.id['x'] > 2

    fits_writer(filename, subset)

    with fits.open(filename) as hdulist:
        assert np.all(np.isnan(hdulist['x'].data[0]))
        assert np.all(np.isnan(hdulist['y'].data[0]))
        np.testing.assert_equal(hdulist['x'].data[1], data['x'][1])
        np.testing.assert_equal(hdulist['y'].data[1], data['y'][1])
        # Here we check BITPIX, not the dtype of the read in data, because if
        # BLANK is present, astropy.io.fits scales the data to float. We want to
        # just make sure here the data is stored with the correct type on disk.
        assert hdulist['x'].header['BITPIX'] == BITPIX[dtype]
        assert hdulist['y'].header['BITPIX'] == BITPIX[dtype]



from __future__ import absolute_import, division, print_function

import numpy as np

from qtpy import QtCore, QtWidgets

from glue.config import colormaps
from glue.viewers.common.qt.toolbar import BasicToolbar
from glue.viewers.matplotlib.qt.widget import MplWidget
from glue.viewers.image.composite_array import CompositeArray
from glue.external.modest_image import imshow
from glue.utils import defer_draw

from glue.viewers.matplotlib.mpl_axes import init_mpl

# Import the mouse mode to make sure it gets registered
from glue.viewers.image.qt.contrast_mouse_mode import ContrastBiasMode  # noqa

__all__ = ['StandaloneImageViewer']


class StandaloneImageViewer(QtWidgets.QMainWindow):
    """
    A simplified image viewer, without any brushing or linking,
    but with the ability to adjust contrast and resample.
    """
    window_closed = QtCore.Signal()
    _toolbar_cls = BasicToolbar
    tools = ['image:contrast', 'image:colormap']

    def __init__(self, image=None, wcs=None, parent=None, **kwargs):
        """
        :param image: Image to display (2D numpy array)
        :param parent: Parent widget (optional)

        :param kwargs: Extra keywords to pass to imshow
        """
        super(StandaloneImageViewer, self).__init__(parent)

        self.central_widget = MplWidget()
        self.setCentralWidget(self.central_widget)
        self._setup_axes()

        self._composite = CompositeArray()
        self._composite.allocate('image')

        self._im = None

        self.initialize_toolbar()

        if image is not None:
            self.set_image(image=image, wcs=wcs, **kwargs)

    def _setup_axes(self):
        _, self._axes = init_mpl(self.central_widget.canvas.fig, axes=None, wcs=True)
        self._axes.set_aspect('equal', adjustable='datalim')

    @defer_draw
    def set_image(self, image=None, wcs=None, **kwargs):
        """
        Update the image shown in the widget
        """
        if self._im is not None:
            self._im.remove()
            self._im = None

        kwargs.setdefault('origin', 'upper')

        if wcs is not None:
            # In the following we force the color and linewith of the WCSAxes
            # frame to be restored after calling reset_wcs. This can be removed
            # once we support Astropy 1.3.1 or later.
            color = self._axes.coords.frame.get_color()
            linewidth = self._axes.coords.frame.get_linewidth()
            self._axes.reset_wcs(wcs)
            self._axes.coords.frame.set_color(color)
            self._axes.coords.frame.set_linewidth(linewidth)
            del color, linewidth

        self._composite.set('image', array=image, color=colormaps.members[0][1])
        self._im = imshow(self._axes, self._composite, **kwargs)
        self._im_array = image
        self._set_norm(self._contrast_mode)

        if 'extent' in kwargs:
            self.axes.set_xlim(kwargs['extent'][:2])
            self.axes.set_ylim(kwargs['extent'][2:])
        else:
            ny, nx = image.shape
            self.axes.set_xlim(-0.5, nx - 0.5)
            self.axes.set_ylim(-0.5, ny - 0.5)

        # FIXME: for a reason I don't quite understand, dataLim doesn't
        # get updated immediately here, which means that there are then
        # issues in the first draw of the image (the limits are such that
        # only part of the image is shown). We just set dataLim manually
        # to avoid this issue. This is also done in ImageViewer.
        self.axes.dataLim.intervalx = self.axes.get_xlim()
        self.axes.dataLim.intervaly = self.axes.get_ylim()

        self._redraw()

    @property
    def axes(self):
        """
        The Matplotlib axes object for this figure
        """
        return self._axes

    def show(self):
        super(StandaloneImageViewer, self).show()
        self._redraw()

    def _redraw(self):
        self.central_widget.canvas.draw()

    def set_cmap(self, cmap):
        self._composite.set('image', color=cmap)
        self._im.invalidate_cache()
        self._redraw()

    def mdi_wrap(self):
        """
        Embed this widget in a GlueMdiSubWindow
        """
        from glue.app.qt.mdi_area import GlueMdiSubWindow
        sub = GlueMdiSubWindow()
        sub.setWidget(self)
        self.destroyed.connect(sub.close)
        self.window_closed.connect(sub.close)
        sub.resize(self.size())
        self._mdi_wrapper = sub
        return sub

    def closeEvent(self, event):
        if self._im is not None:
            self._im.remove()
            self._im = None
        self.window_closed.emit()
        return super(StandaloneImageViewer, self).closeEvent(event)

    def _set_norm(self, mode):
        """
        Use the `ContrastMouseMode` to adjust the transfer function
        """

        pmin, pmax = mode.get_clip_percentile()

        if pmin is None:
            clim = mode.get_vmin_vmax()
        else:
            clim = (np.nanpercentile(self._im_array, pmin),
                    np.nanpercentile(self._im_array, pmax))

        stretch = mode.stretch
        self._composite.set('image', clim=clim, stretch=stretch,
                            bias=mode.bias, contrast=mode.contrast)

        self._im.invalidate_cache()
        self._redraw()

    def initialize_toolbar(self):

        from glue.config import viewer_tool

        self.toolbar = self._toolbar_cls(self)

        for tool_id in self.tools:
            mode_cls = viewer_tool.members[tool_id]
            if tool_id == 'image:contrast':
                mode = mode_cls(self, move_callback=self._set_norm)
                self._contrast_mode = mode
            else:
                mode = mode_cls(self)
            self.toolbar.add_tool(mode)

        self.addToolBar(self.toolbar)

    def set_status(self, message):
        sb = self.statusBar()
        sb.showMessage(message)



from __future__ import absolute_import, division, print_function

from glue.utils import defer_draw, decorate_all_methods
from glue.viewers.matplotlib.qt.data_viewer import MatplotlibDataViewer
from glue.viewers.scatter.qt.layer_style_editor import ScatterLayerStyleEditor
from glue.viewers.scatter.layer_artist import ScatterLayerArtist
from glue.viewers.scatter.qt.options_widget import ScatterOptionsWidget
from glue.viewers.scatter.state import ScatterViewerState

from glue.viewers.scatter.viewer import MatplotlibScatterMixin

__all__ = ['ScatterViewer']


@decorate_all_methods(defer_draw)
class ScatterViewer(MatplotlibScatterMixin, MatplotlibDataViewer):

    LABEL = '2D Scatter'
    _layer_style_widget_cls = ScatterLayerStyleEditor
    _state_cls = ScatterViewerState
    _options_cls = ScatterOptionsWidget
    _data_artist_cls = ScatterLayerArtist
    _subset_artist_cls = ScatterLayerArtist

    tools = ['select:rectangle', 'select:xrange',
             'select:yrange', 'select:circle',
             'select:polygon']

    def __init__(self, session, parent=None, state=None):
        MatplotlibDataViewer.__init__(self, session, parent=parent, state=state)
        MatplotlibScatterMixin.setup_callbacks(self)



from __future__ import absolute_import, division, print_function

from collections import OrderedDict

from mock import MagicMock
import pytest
import numpy as np
from numpy.testing import assert_equal
from glue.core import DataCollection, Data

from ..subset_mask import SubsetMaskImporter, SubsetMaskExporter


class MySubsetMaskImporter(SubsetMaskImporter):

    filename = None
    reader = None

    def get_filename_and_reader(self):
        return self.filename, self.reader


class MySubsetMaskExporter(SubsetMaskExporter):

    filename = None
    writer = None

    def get_filename_and_writer(self):
        return self.filename, self.writer


class TestImporter():

    def setup_method(self, method):
        self.importer = MySubsetMaskImporter()
        self.importer.filename = 'test-filename'
        self.importer.reader = MagicMock()
        self.data = Data(x=[1, 2, 3])
        self.data_collection = DataCollection([self.data])

    def test_single_valid(self):
        self.importer.reader.return_value = OrderedDict([('subset 1', np.array([0, 1, 0]))])
        self.importer.run(self.data, self.data_collection)
        assert len(self.data_collection.subset_groups) == 1
        assert_equal(self.data.subsets[0].to_mask(), [0, 1, 0])

    def test_multiple_valid(self):
        self.importer.reader.return_value = OrderedDict([('subset 1', np.array([0, 1, 0])),
                                                         ('subset 2', np.array([1, 1, 0]))])
        self.importer.run(self.data, self.data_collection)
        assert len(self.data_collection.subset_groups) == 2
        assert_equal(self.data.subsets[0].to_mask(), [0, 1, 0])
        assert_equal(self.data.subsets[1].to_mask(), [1, 1, 0])

    def test_missing_masks(self):
        self.importer.reader.return_value = OrderedDict()
        with pytest.raises(ValueError) as exc:
            self.importer.run(self.data, self.data_collection)
        assert exc.value.args[0] == "No subset masks were returned"

    def test_single_invalid_shape(self):
        self.importer.reader.return_value = OrderedDict([('subset 1', np.array([0, 1, 0, 1]))])
        with pytest.raises(ValueError) as exc:
            self.importer.run(self.data, self.data_collection)
        assert exc.value.args[0].replace('L', '') == "Mask shape (4,) does not match data shape (3,)"

    def test_multiple_inconsistent_shapes(self):
        self.importer.reader.return_value = OrderedDict([('subset 1', np.array([0, 1, 0])),
                                                         ('subset 2', np.array([0, 1, 0, 1]))])
        with pytest.raises(ValueError) as exc:
            self.importer.run(self.data, self.data_collection)
        assert exc.value.args[0] == "Not all subsets have the same shape"

    def test_subset_single(self):
        self.importer.reader.return_value = OrderedDict([('subset 1', np.array([0, 1, 0]))])
        subset = self.data.new_subset()
        assert_equal(self.data.subsets[0].to_mask(), [0, 0, 0])
        self.importer.run(subset, self.data_collection)
        assert_equal(self.data.subsets[0].to_mask(), [0, 1, 0])

    def test_subset_multiple(self):
        self.importer.reader.return_value = OrderedDict([('subset 1', np.array([0, 1, 0])),
                                                         ('subset 2', np.array([1, 1, 0]))])
        subset = self.data.new_subset()
        with pytest.raises(ValueError) as exc:
            self.importer.run(subset, self.data_collection)
        assert exc.value.args[0] == 'Can only read in a single subset when importing into a subset'


class TestExporter():

    def setup_method(self, method):
        self.exporter = MySubsetMaskExporter()
        self.exporter.filename = 'test-filename'
        self.exporter.writer = MagicMock()
        self.data = Data(x=[1, 2, 3])
        self.data_collection = DataCollection([self.data])

    def test_no_subsets(self):
        with pytest.raises(ValueError) as exc:
            self.exporter.run(self.data)
        assert exc.value.args[0] == 'Data has no subsets'

    def test_multiple_valid(self):
        self.subset1 = self.data_collection.new_subset_group(subset_state=self.data.id['x'] >= 2,
                                                             label='subset a')
        self.subset2 = self.data_collection.new_subset_group(subset_state=self.data.id['x'] >= 3,
                                                             label='subset b')
        self.exporter.run(self.data)
        assert self.exporter.writer.call_count == 1
        assert self.exporter.writer.call_args[0][0] == 'test-filename'
        masks = self.exporter.writer.call_args[0][1]
        assert len(masks) == 2
        assert_equal(masks['subset a'], [0, 1, 1])
        assert_equal(masks['subset b'], [0, 0, 1])

    def test_single_subset_valid(self):
        self.subset = self.data_collection.new_subset_group(subset_state=self.data.id['x'] >= 2,
                                                            label='subset a')
        self.exporter.run(self.data.subsets[0])
        assert self.exporter.writer.call_count == 1
        assert self.exporter.writer.call_args[0][0] == 'test-filename'
        masks = self.exporter.writer.call_args[0][1]
        assert len(masks) == 1
        assert_equal(masks['subset a'], [0, 1, 1])



from __future__ import absolute_import, division, print_function

import logging
from contextlib import contextmanager
from weakref import WeakKeyDictionary
from inspect import getmro
from collections import Counter

from glue.core.exceptions import InvalidSubscriber, InvalidMessage
from glue.core.message import Message
from glue.core.hub_callback_container import HubCallbackContainer

__all__ = ['Hub', 'HubListener']


class Hub(object):

    """The hub manages communication between subscribers.

    Objects :func:`subscribe` to receive specific message types. When
    a message is passed to :func:`broadcast`, the hub observes the
    following protocol:

        * For each subscriber, it looks for a message class
          subscription that is a superclass of the input message type
          (if several are found, the most-subclassed one is chosen)

        * If one is found, it calls the subscriptions filter(message)
          class (if provided)

        * If filter(message) == True, it calls handler(message)
          (or notify(message) if handler wasn't provided).

    """

    def __init__(self, *args):
        """
        Any arguments that are passed to Hub will be registered
        to the new hub object.
        """
        # Dictionary of subscriptions
        self._subscriptions = WeakKeyDictionary()

        self._paused = False
        self._queue = []

        self._ignore = Counter()

        from glue.core.data import BaseData
        from glue.core.subset import Subset
        from glue.core.data_collection import DataCollection

        listeners = set(filter(lambda x: isinstance(x, HubListener), args))
        data = set(filter(lambda x: isinstance(x, BaseData), args))
        subsets = set(filter(lambda x: isinstance(x, Subset), args))
        dcs = set(filter(lambda x: isinstance(x, DataCollection), args))
        listeners -= (data | subsets | dcs)
        if set(listeners | data | subsets | dcs) != set(args):
            raise TypeError("Inputs must be HubListener, data, subset, or "
                            "data collection objects")

        for l in listeners:
            l.register_to_hub(self)
        for d in data:
            d.register_to_hub(self)
        for dc in dcs:
            dc.register_to_hub(self)
        for s in subsets:
            s.register()

    def subscribe(self, subscriber, message_class,
                  handler=None,
                  filter=lambda x: True):
        """Subscribe an object to a type of message class.

        :param subscriber: The subscribing object
        :type subscriber: :class:`~glue.core.hub.HubListener`

        :param message_class: A :class:`~glue.core.message.Message` class
                              to subscribe to

        :param handler:
           An optional function of the form handler(message) that will
           receive the message on behalf of the subscriber. If not provided,
           this defaults to the HubListener's notify method


        :param filter:
           An optional function of the form filter(message). Messages
           are only passed to the subscriber if filter(message) == True.
           The default is to always pass messages.


        Raises:
            InvalidMessage: If the input class isn't a
            :class:`~glue.core.message.Message` class

            InvalidSubscriber: If the input subscriber isn't a
            HubListener object.

        """
        if not isinstance(subscriber, HubListener):
            raise InvalidSubscriber("Subscriber must be a HubListener: %s" %
                                    type(subscriber))
        if not isinstance(message_class, type) or \
                not issubclass(message_class, Message):
            raise InvalidMessage("message class must be a subclass of "
                                 "glue.Message: %s" % type(message_class))
        logging.getLogger(__name__).info("Subscribing %s to %s",
                                         subscriber, message_class.__name__)

        if not handler:
            handler = subscriber.notify

        if subscriber not in self._subscriptions:
            self._subscriptions[subscriber] = HubCallbackContainer()

        self._subscriptions[subscriber][message_class] = handler, filter

    def is_subscribed(self, subscriber, message):
        """
        Test whether the subscriber has subscribed to a given message class

        :param subscriber: The subscriber to test
        :param message: The message class to test

        Returns:

            True if the subscriber/message pair have been subscribed to the hub

        """
        return (subscriber in self._subscriptions and
                message in self._subscriptions[subscriber])

    def get_handler(self, subscriber, message):
        if subscriber is None:
            return None
        try:
            return self._subscriptions[subscriber][message][0]
        except KeyError:
            return None

    def unsubscribe(self, subscriber, message):
        """
        Remove a (subscriber,message) pair from subscription list.
        The handler originally attached to the subscription will
        no longer be called when broadcasting messages of type message
        """
        if subscriber not in self._subscriptions:
            return
        if message in self._subscriptions[subscriber]:
            self._subscriptions[subscriber].pop(message)

    def unsubscribe_all(self, subscriber):
        """
        Unsubscribe the object from any subscriptions.
        """
        if subscriber in self._subscriptions:
            self._subscriptions.pop(subscriber)

    def _find_handlers(self, message):
        """Yields all (subscriber, handler) pairs that should receive a message
        """
        # self._subscriptions:
        # subscriber => { message type => (filter, handler)}

        # loop over subscribed objects
        for subscriber, subscriptions in list(self._subscriptions.items()):

            # subscriptions to message or its superclasses
            messages = [msg for msg in subscriptions.keys() if
                        issubclass(type(message), msg)]

            if len(messages) == 0:
                continue

            # narrow to the most-specific message
            candidate = max(messages, key=_mro_count)

            handler, test = subscriptions[candidate]
            if test(message):
                yield subscriber, handler

    @contextmanager
    def ignore_callbacks(self, ignore_type):
        self._ignore[ignore_type] += 1
        try:
            yield
        finally:
            self._ignore[ignore_type] -= 1

    @contextmanager
    def delay_callbacks(self):
        self._paused = True
        try:
            yield
        finally:
            self._paused = False
            # TODO: could de-duplicate messages here
            for message in self._queue:
                self.broadcast(message)
            self._queue = []

    def broadcast(self, message):
        """Broadcasts a message to all subscribed objects.

        :param message: The message to broadcast
        :type message: :class:`~glue.core.message.Message`
        """
        if self._ignore.get(type(message), 0) > 0:
            return
        elif self._paused:
            self._queue.append(message)
        else:
            logging.getLogger(__name__).info("Broadcasting %s", message)
            for subscriber, handler in self._find_handlers(message):
                handler(message)

    def __getstate__(self):
        """ Return a picklable representation of the hub

        Note: Only objects in glue.core are currently supported
        as pickleable. Thus, any subscriptions from objects outside
        glue.core will note be saved or restored
        """
        result = self.__dict__.copy()
        result['_subscriptions'] = self._subscriptions.copy()
        for s in self._subscriptions:
            try:
                module = s.__module__
            except AttributeError:
                module = ''
            if not module.startswith('glue.core'):
                print('Pickle warning: Hub removing subscription to %s' % s)
                result['_subscriptions'].pop(s)
        return result


class HubListener(object):

    """
    The base class for any object that subscribes to hub messages.
    This interface defines a single method, notify, that receives
    messages
    """

    def register_to_hub(self, hub):
        raise NotImplementedError

    def unregister(self, hub):
        """ Default unregistration action. Calls hub.unsubscribe_all on self"""
        hub.unsubscribe_all(self)

    def notify(self, message):
        raise NotImplementedError("Message has no handler: %s" % message)


def _mro_count(obj):
    return len(getmro(obj))



from __future__ import absolute_import, division, print_function

import os
from glue.qglue import parse_data
from glue.tests.helpers import requires_spectral_cube

DATA = os.path.join(os.path.dirname(__file__), 'data')


@requires_spectral_cube
def test_identifier():
    from ..spectral_cube import is_spectral_cube
    assert is_spectral_cube(os.path.join(DATA, 'cube_3d.fits'))


@requires_spectral_cube
def test_reader():
    from ..spectral_cube import read_spectral_cube
    data = read_spectral_cube(os.path.join(DATA, 'cube_3d.fits'))
    data['STOKES I']
    assert data.shape == (2, 3, 4)


@requires_spectral_cube
def test_qglue():
    from spectral_cube import SpectralCube
    cube = SpectralCube.read(os.path.join(DATA, 'cube_3d.fits'))
    data = parse_data(cube, 'x')[0]
    assert data.label == 'x'
    data['STOKES I']
    assert data.shape == (2, 3, 4)



def setup():

    from glue.config import exporters
    from glue.plugins.exporters.plotly.export_plotly import (can_save_plotly, DISPATCH,
                                                             export_scatter, export_histogram)
    from glue.plugins.exporters.plotly.qt.exporter import save_plotly

    from glue.viewers.scatter.qt import ScatterViewer
    from glue.viewers.histogram.qt import HistogramViewer

    DISPATCH[ScatterViewer] = export_scatter
    DISPATCH[HistogramViewer] = export_histogram

    exporters.add('Plotly', save_plotly, can_save_plotly)



import numpy as np
from glue.core.exceptions import IncompatibleAttribute

__all__ = ['get_mask_with_key_joins']


def concatenate_arrays(*arrays):
    """
    Given N arrays of size M, return an array of size M where each item is
    the concatenation of the bytes of the orginal arrays for the respective
    item.
    """

    # Find the total size of an item in the final array
    total_size = sum(x.dtype.itemsize for x in arrays)

    # Set up the final array
    buffer = np.zeros(len(arrays[0]) * total_size, dtype=np.byte)

    # Now view this array as a structured array
    colnames = ['{0:x}'.format(x) for x in range(len(arrays))]
    dtype = list(zip(colnames, [x.dtype for x in arrays]))
    buffer_as_struct = buffer.view(dtype)

    # Finally, fill the array column by column
    for letter, array in zip(colnames, arrays):
        buffer_as_struct[letter] = array

    # And view as an array of size M
    return buffer_as_struct.view('S{0}'.format(total_size))


def get_mask_with_key_joins(data, key_joins, subset_state, view=None):
    """
    Given a dataset and a subset state, check whether the subset state
    can be translated to the current dataset via key joins.

    Note that this does not try simply applying the subset state to the
    dataset, as it is assumed this has been tried first.
    """

    for other, (cid1, cid2) in key_joins.items():

        if getattr(other, '_recursing', False):
            continue

        try:
            data._recursing = True
            mask_right = other.get_mask(subset_state)
        except IncompatibleAttribute:
            continue
        finally:
            data._recursing = False

        if len(cid1) == 1 and len(cid2) == 1:

            key_left = data.get_data(cid1[0], view=view)
            key_right = other.get_data(cid2[0], view=mask_right)
            mask = np.in1d(key_left.ravel(), key_right.ravel())

            return mask.reshape(key_left.shape)

        elif len(cid1) == len(cid2):

            key_left_all = []
            key_right_all = []

            for cid1_i, cid2_i in zip(cid1, cid2):
                key_left_all.append(data.get_data(cid1_i, view=view).ravel())
                key_right_all.append(other.get_data(cid2_i, view=mask_right).ravel())

            key_left_all = concatenate_arrays(*key_left_all)
            key_right_all = concatenate_arrays(*key_right_all)

            mask = np.in1d(key_left_all, key_right_all)

            return mask.reshape(data.get_data(cid1_i, view=view).shape)

        elif len(cid1) == 1:

            key_left = data.get_data(cid1[0], view=view).ravel()
            mask = np.zeros_like(key_left, dtype=bool)
            for cid2_i in cid2:
                key_right = other.get_data(cid2_i, view=mask_right).ravel()
                mask |= np.in1d(key_left, key_right)

            return mask.reshape(data.get_data(cid1[0], view=view).shape)

        elif len(cid2) == 1:

            key_right = other.get_data(cid2[0], view=mask_right).ravel()
            mask = np.zeros_like(data.get_data(cid1[0], view=view).ravel(), dtype=bool)
            for cid1_i in cid1:
                key_left = data.get_data(cid1_i, view=view).ravel()
                mask |= np.in1d(key_left, key_right)

            return mask.reshape(data.get_data(cid1[0], view=view).shape)

        else:

            raise Exception("Either the number of components in the key join sets "
                            "should match, or one of the component sets should ",
                            "contain a single component.")

    raise IncompatibleAttribute



from __future__ import absolute_import, division, print_function

import numpy as np

from glue.core.data_factories.helpers import has_extension
from glue.core.data import Component, Data
from glue.config import data_factory, qglue_parser


__all__ = ['astropy_tabular_data', 'sextractor_factory', 'cds_factory',
           'daophot_factory', 'ipac_factory', 'aastex_factory',
           'latex_factory']


# In this file, we define data factories based on the Astropy table reader.


def is_readable_by_astropy(filename, **kwargs):
    # This identifier is not efficient, because it involves actually trying
    # to read in the table. However, we only use this as the identifier for
    # the astropy_tabular_data factory which has a priority of 0 and is
    # therefore only used as a last attempt if all else fails.
    try:
        astropy_table_read(filename, **kwargs)
    except Exception:
        return False
    else:
        return True


def astropy_table_read(*args, **kwargs):

    from astropy.table import Table

    # In Python 3, as of Astropy 0.4, if the format is not specified, the
    # automatic format identification will fail (astropy/astropy#3013).
    # This is only a problem for ASCII formats however, because it is due
    # to the fact that the file object in io.ascii does not rewind to the
    # start between guesses (due to a bug), so here we can explicitly try
    # the ASCII format if the format keyword was not already present. But
    # also more generally, we should first try the ASCII readers.
    if 'format' not in kwargs:
        try:
            return Table.read(*args, format='ascii', **kwargs)
        except Exception:
            pass

    # If the above didn't work, attempt to read with no specified format
    return Table.read(*args, **kwargs)


@data_factory(label="Catalog (astropy.table parser)",
              identifier=is_readable_by_astropy,
              priority=0)
def astropy_tabular_data(*args, **kwargs):
    """
    Build a data set from a table. We restrict ourselves to tables
    with 1D columns.

    All arguments are passed to
        astropy.table.Table.read(...).
    """

    result = Data()

    table = astropy_table_read(*args, **kwargs)

    result.meta = table.meta

    # Loop through columns and make component list
    for column_name in table.columns:
        c = table[column_name]
        u = c.unit if hasattr(c, 'unit') else c.units

        if table.masked:
            # fill array for now
            try:
                c = c.filled(fill_value=np.nan)
            except (ValueError, TypeError):  # assigning nan to integer dtype
                c = c.filled(fill_value=-1)

        nc = Component.autotyped(c, units=u)
        result.add_component(nc, column_name)

    return result


@data_factory(label="VO table",
              identifier=has_extension('xml vot xml.gz vot.gz'),
              priority=1)
def astropy_tabular_data_votable(*args, **kwargs):
    kwargs['format'] = 'votable'
    return astropy_tabular_data(*args, **kwargs)


@data_factory(label="FITS table",
              identifier=has_extension('fits fits.gz fit fit.gz'),
              priority=1)
def astropy_tabular_data_fits(*args, **kwargs):
    kwargs['format'] = 'fits'
    return astropy_tabular_data(*args, **kwargs)


# Add explicit factories for the formats which astropy.table
# can parse, but does not auto-identify


def formatted_table_factory(format, label):

    @data_factory(label=label, identifier=lambda *a, **k: False)
    def factory(file, **kwargs):
        kwargs['format'] = 'ascii.%s' % format
        return astropy_tabular_data(file, **kwargs)

    # rename function to its variable reference below
    # allows pickling to work
    factory.__name__ = '%s_factory' % format

    return factory


sextractor_factory = formatted_table_factory('sextractor', 'SExtractor Catalog')
cds_factory = formatted_table_factory('cds', 'CDS Catalog')
daophot_factory = formatted_table_factory('daophot', 'DAOphot Catalog')
ipac_factory = formatted_table_factory('ipac', 'IPAC Catalog')
aastex_factory = formatted_table_factory('aastex', 'AASTeX Table')
latex_factory = formatted_table_factory('latex', 'LaTeX Table')

try:
    from astropy.table import Table
except ImportError:
    pass
else:
    @qglue_parser(Table)
    def _parse_data_astropy_table(data, label):
        kwargs = dict((c, data[c]) for c in data.columns)
        return [Data(label=label, **kwargs)]



from numpy.testing import assert_allclose

import matplotlib.pyplot as plt

from glue.core.application_base import Application
from glue.viewers.common.viewer import Viewer
from glue.viewers.matplotlib.viewer import MatplotlibViewerMixin
from glue.viewers.matplotlib.state import MatplotlibDataViewerState


def assert_limits(viewer, x_min, x_max, y_min, y_max):
    # Convenience to check both state and matplotlib
    assert_allclose(viewer.state.x_min, x_min)
    assert_allclose(viewer.state.x_max, x_max)
    assert_allclose(viewer.state.y_min, y_min)
    assert_allclose(viewer.state.y_max, y_max)
    assert_allclose(viewer.axes.get_xlim(), (x_min, x_max))
    assert_allclose(viewer.axes.get_ylim(), (y_min, y_max))


def test_aspect_ratio():

    # Test of the aspect ratio infrastructure

    class CustomViewer(MatplotlibViewerMixin, Viewer):

        _state_cls = MatplotlibDataViewerState

        def __init__(self, *args, **kwargs):
            Viewer.__init__(self, *args, **kwargs)
            self.figure = plt.figure(figsize=(12, 6))
            self.axes = self.figure.add_axes([0, 0, 1, 1])
            MatplotlibViewerMixin.setup_callbacks(self)

        def show(self):
            pass

    class CustomApplication(Application):
        def add_widget(self, *args, **kwargs):
            pass

    app = CustomApplication()

    viewer = app.new_data_viewer(CustomViewer)
    viewer.state.aspect = 'equal'

    assert_limits(viewer, -0.5, 1.5, 0., 1.)

    # Test changing x limits in state, which should just change the y limits

    viewer.state.x_min = -2.5
    assert_limits(viewer, -2.5, 1.5, -0.5, 1.5)

    viewer.state.x_max = -1.5
    assert_limits(viewer, -2.5, -1.5, 0.25, 0.75)

    # Test changing y limits in state, which should just change the x limits

    viewer.state.y_max = 1.25
    assert_limits(viewer, -3.0, -1.0, 0.25, 1.25)

    viewer.state.y_min = 0.75
    assert_limits(viewer, -2.5, -1.5, 0.75, 1.25)

    # Test changing x limits in Matplotlib, which should just change the x limits

    viewer.axes.set_xlim(1., 3.)
    assert_limits(viewer, 1.0, 3.0, 0.5, 1.5)

    # Test changing x limits in Matplotlib, which should just change the x limits

    viewer.axes.set_ylim(0., 2.)
    assert_limits(viewer, 0.0, 4.0, 0.0, 2.0)

    # We include tests for resizing inside the Qt folder (since this doesn't
    # work correctly with the Agg backend)



from __future__ import absolute_import, division, print_function

import os
import sys
import logging
import traceback
import webbrowser

from qtpy import QtWidgets
from glue.utils import nonpartial
from glue.utils.qt import load_ui
from glue.utils.qt.widget_properties import TextProperty, ButtonProperty
from glue.plugins.exporters.plotly.export_plotly import build_plotly_call


def save_plotly(application):
    """
    Save a Glue session to a plotly plot

    This is currently restricted to 1-4 scatterplots or histograms

    Parameters
    ----------
    application : `~glue.core.application_base.Application`
        Glue application to save
    label : str
        Label for the exported plot
    """

    args, kwargs = build_plotly_call(application)

    logging.getLogger(__name__).debug(args, kwargs)

    # TODO: check what current GUI framework is

    exporter = QtPlotlyExporter(plotly_args=args, plotly_kwargs=kwargs)
    exporter.exec_()


class QtPlotlyExporter(QtWidgets.QDialog):

    save_settings = ButtonProperty('checkbox_save')
    username = TextProperty('text_username')
    api_key = TextProperty('text_api_key')
    title = TextProperty('text_title')
    legend = ButtonProperty('checkbox_legend')

    def __init__(self, plotly_args=[], plotly_kwargs={}, parent=None):

        super(QtPlotlyExporter, self).__init__(parent=parent)

        self.plotly_args = plotly_args
        self.plotly_kwargs = plotly_kwargs

        self.ui = load_ui('exporter.ui', self,
                          directory=os.path.dirname(__file__))

        self.button_cancel.clicked.connect(self.reject)
        self.button_export.clicked.connect(self.accept)

        # Set up radio button groups

        self._radio_account = QtWidgets.QButtonGroup()
        self._radio_account.addButton(self.ui.radio_account_config)
        self._radio_account.addButton(self.ui.radio_account_manual)

        self._radio_sharing = QtWidgets.QButtonGroup()
        self._radio_sharing.addButton(self.ui.radio_sharing_public)
        self._radio_sharing.addButton(self.ui.radio_sharing_secret)
        self._radio_sharing.addButton(self.ui.radio_sharing_private)

        # Find out stored credentials (note that this will create the
        # credentials file if it doesn't already exist)

        from plotly import plotly

        credentials = plotly.get_credentials()
        config_available = credentials['username'] != "" and credentials['api_key'] != ""

        if config_available:
            self.ui.radio_account_config.setChecked(True)
            label = self.ui.radio_account_config.text()
            self.ui.radio_account_config.setText(label + " (username: {0})".format(credentials['username']))
        else:
            self.ui.radio_account_manual.setChecked(True)

        self.ui.radio_sharing_secret.setChecked(True)

        self.ui.text_username.textChanged.connect(nonpartial(self._set_manual_mode))
        self.ui.text_api_key.textChanged.connect(nonpartial(self._set_manual_mode))

        self.set_status('', color='black')

    def _set_manual_mode(self):
        self.ui.radio_account_manual.setChecked(True)

    def accept(self):

        # In future we might be able to use more fine-grained exceptions
        # https://github.com/plotly/plotly.py/issues/524

        self.set_status('Signing in and plotting...', color='blue')

        auth = {}

        if self.ui.radio_account_config.isChecked():
            auth['username'] = ''
            auth['api_key'] = ''
        else:
            if self.username == "":
                self.set_status("Username not set", color='red')
                return
            elif self.api_key == "":
                self.set_status("API key not set", color='red')
                return
            else:
                auth['username'] = self.username
                auth['api_key'] = self.api_key

        from plotly import plotly
        from plotly.exceptions import PlotlyError
        from plotly.tools import set_credentials_file

        if self.ui.radio_sharing_public.isChecked():
            self.plotly_kwargs['sharing'] = 'public'
        elif self.ui.radio_sharing_secret.isChecked():
            self.plotly_kwargs['sharing'] = 'secret'
        else:
            self.plotly_kwargs['sharing'] = 'private'

        # We need to fix URLs, so we can't let plotly open it yet
        # https://github.com/plotly/plotly.py/issues/526
        self.plotly_kwargs['auto_open'] = False

        # Get title and legend preferences from the window
        self.plotly_args[0]['layout']['showlegend'] = self.legend
        self.plotly_args[0]['layout']['title'] = self.title

        try:
            plotly.sign_in(auth['username'], auth['api_key'])
            plotly_url = plotly.plot(*self.plotly_args, **self.plotly_kwargs)
        except PlotlyError as exc:
            print("Plotly exception:")
            print('-' * 60)
            traceback.print_exc(file=sys.stdout)
            print('-' * 60)
            if ('the supplied API key doesn\'t match our records' in exc.args[0] or
                    'Sign in failed' in exc.args[0]):
                username = auth['username'] or plotly.get_credentials()['username']
                self.set_status("Authentication failed".format(username), color='red')
            elif "filled its quota of private files" in exc.args[0]:
                self.set_status("Maximum number of private plots reached", color='red')
            elif "Accounts on the Community Plan cannot save private files" in exc.args[0]:
                self.set_status("Accounts on the Community Plan cannot save private files", color='red')
            else:
                self.set_status("An unexpected error occurred", color='red')
            return
        except Exception:
            print("Plotly exception:")
            print('-' * 60)
            traceback.print_exc(file=sys.stdout)
            print('-' * 60)
            self.set_status("An unexpected error occurred", color='red')
            return

        self.set_status('Exporting succeeded', color='blue')

        if self.save_settings and self.ui.radio_account_manual.isChecked():
            try:
                set_credentials_file(**auth)
            except Exception:
                print("Plotly exception:")
                print('-' * 60)
                traceback.print_exc(file=sys.stdout)
                print('-' * 60)
                self.set_status('Exporting succeeded (but saving login failed)', color='blue')

        print("Plotly URL: {0}".format(plotly_url))

        webbrowser.open_new_tab(plotly_url)

        super(QtPlotlyExporter, self).accept()

    def set_status(self, text, color):
        self.ui.text_status.setText(text)
        self.ui.text_status.setStyleSheet("color: {0}".format(color))
        QtWidgets.QApplication.instance().processEvents()



from __future__ import absolute_import, division, print_function

import numpy as np

try:
    from plotly import plotly
except ImportError:
    plotly = None

from glue.core.layout import Rectangle, snap_to_grid
from glue.utils import categorical_ndarray

SYM = {'o': 'circle', 's': 'square', '+': 'cross', '^': 'triangle-up',
       '*': 'cross'}


def _sanitize(*arrs):
    mask = np.ones(arrs[0].shape, dtype=np.bool)
    for a in arrs:
        try:
            mask &= (~np.isnan(a))
        except TypeError:  # non-numeric dtype
            pass

    return tuple(a[mask].ravel() for a in arrs)


def _position_plots(viewers, layout):
    rs = [Rectangle(v.position[0], v.position[1],
                    v.viewer_size[0], v.viewer_size[1])
          for v in viewers]
    right = max(r.x + r.w for r in rs)
    top = max(r.y + r.h for r in rs)
    for r in rs:
        r.x = 1. * r.x / right
        r.y = 1. - 1. * (r.y + r.h) / top
        r.w = 1. * r.w / right
        r.h = 1. * r.h / top

    grid = snap_to_grid(rs, padding=0.05)
    grid = dict((v, grid[r]) for v, r in zip(viewers, rs))

    for i, plot in enumerate(viewers, 1):
        g = grid[plot]
        xdomain = [g.x, g.x + g.w]
        ydomain = [g.y, g.y + g.h]
        suffix = '' if i == 1 else str(i)

        xax, yax = 'xaxis' + suffix, 'yaxis' + suffix
        layout[xax].update(domain=xdomain, anchor=yax.replace('axis', ''))
        layout[yax].update(domain=ydomain, anchor=xax.replace('axis', ''))


def _stack_horizontal(layout):
    layout['xaxis']['domain'] = [0, 0.45]
    layout['xaxis2']['domain'] = [0.55, 1]
    layout['yaxis2']['anchor'] = 'x2'


def _grid_2x23(layout):
    opts = {
        'xaxis': {'domain': [0, 0.45]},
        'yaxis': {'domain': [0, 0.45]},
        'xaxis2': {"domain": [0.55, 1]},
        'yaxis2': {"domain": [0, 0.45],
                   "anchor": "x2"
                   },
        'xaxis3': {
            "domain": [0, 0.45],
            "anchor": "y3"
        },
        'yaxis3': {
            "domain": [0.55, 1],
        },
        'xaxis4': {
            "domain": [0.55, 1],
            "anchor": "y4",
        },
        'yaxis4': {
            "domain": [0.55, 1],
            "anchor": "x4"
        }
    }
    for k, v in opts.items():
        if k not in layout:
            continue
        layout[k].update(**v)


def _axis(log=False, lo=0, hi=1, title='', categorical=False):
    if log:
        if lo < 0:
            lo = 1e-3
        if hi < 0:
            hi = 1e-3
        lo = np.log10(lo)
        hi = np.log10(hi)

    result = dict(type='log' if log else 'linear',
                  rangemode='normal',
                  range=[lo, hi], title=title)

    return result


def _fix_legend_duplicates(traces, layout):
    """Prevent repeat entries in the legend"""
    seen = set()
    for t in traces:
        key = (t.get('name'), t.get('marker', {}).get('color'))
        if key in seen:
            t['showlegend'] = False
        else:
            seen.add(key)


def _color(style):
    r, g, b, a = style.rgba
    r = int(r * 255)
    g = int(g * 255)
    b = int(b * 255)
    return 'rgba(%i, %i, %i, %0.1f)' % (r, g, b, a)


def export_scatter(viewer):
    """Export a scatter viewer to a list of
    plotly-formatted data dictionaries"""
    traces = []
    xatt, yatt = viewer.state.x_att, viewer.state.y_att
    xcat = ycat = False

    for layer in viewer.layers:
        if not layer.visible:
            continue
        l = layer.layer
        xcat |= l.data.get_kind(xatt) == 'categorical'
        ycat |= l.data.get_kind(yatt) == 'categorical'

        marker = dict(symbol=SYM.get(l.style.marker, 'circle'),
                      color=_color(l.style),
                      size=l.style.markersize)

        x, y = l[xatt], l[yatt]
        if isinstance(x, categorical_ndarray):
            x = x.codes
        if isinstance(y, categorical_ndarray):
            y = y.codes

        x, y = _sanitize(x, y)

        trace = dict(x=x, y=y,
                     type='scatter',
                     mode='markers',
                     marker=marker,
                     name=l.label)

        traces.append(trace)

    xaxis = _axis(log=viewer.state.x_log, lo=viewer.state.x_min, hi=viewer.state.x_max,
                  title=viewer.state.x_att.label, categorical=xcat)
    yaxis = _axis(log=viewer.state.y_log, lo=viewer.state.y_min, hi=viewer.state.y_max,
                  title=viewer.state.y_att.label, categorical=ycat)

    return traces, xaxis, yaxis


def export_histogram(viewer):
    traces = []
    att = viewer.state.x_att
    ymax = 1e-3
    for artist in viewer.layers:
        if not artist.visible:
            continue
        artist.wait()
        layer = artist.layer
        edges, hist = artist.state.histogram
        x, y = _sanitize(edges[:-1], hist)
        trace = dict(
            name=layer.label,
            type='bar',
            marker=dict(color=_color(layer.style)),
            x=x,
            y=y)
        traces.append(trace)
        ymax = max(ymax, hist.max())

    xlabel = att.label
    xmin, xmax = viewer.state.x_min, viewer.state.x_max
    if viewer.state.x_log:
        xlabel = 'Log ' + xlabel
        xmin = np.log10(xmin)
        xmax = np.log10(xmax)
    xaxis = _axis(lo=xmin, hi=xmax, title=xlabel)
    yaxis = _axis(log=viewer.state.y_log, lo=0 if not viewer.state.y_log else 1e-3,
                  hi=ymax * 1.05)

    return traces, xaxis, yaxis


def build_plotly_call(app):
    args = []
    layout = {'showlegend': True, 'barmode': 'overlay',
              'title': 'Autogenerated by Glue'}

    ct = 1
    for tab in app.viewers:
        for viewer in tab:
            if hasattr(viewer, '__plotly__'):
                p, xaxis, yaxis = viewer.__plotly__()
            else:
                assert type(viewer) in DISPATCH
                p, xaxis, yaxis = DISPATCH[type(viewer)](viewer)

            xaxis['zeroline'] = False
            yaxis['zeroline'] = False

            suffix = '' if ct == 1 else '%i' % ct
            layout['xaxis' + suffix] = xaxis
            layout['yaxis' + suffix] = yaxis
            if ct > 1:
                yaxis['anchor'] = 'x' + suffix
                for item in p:
                    item['xaxis'] = 'x' + suffix
                    item['yaxis'] = 'y' + suffix
            ct += 1
            args.extend(p)

    _position_plots([v for tab in app.viewers for v in tab], layout)
    _fix_legend_duplicates(args, layout)

    return [dict(data=args, layout=layout)], {}


def can_save_plotly(application):
    """
    Check whether an application can be exported to plotly

    Raises an exception if not
    """
    if not plotly:
        raise ValueError("Plotly Export requires the plotly python library. "
                         "Please install first")

    for tab in application.viewers:
        for viewer in tab:
            if hasattr(viewer, '__plotly__'):
                continue

            if not isinstance(viewer, tuple(DISPATCH)):
                raise ValueError("Plotly Export cannot handle viewer: %s"
                                 % type(viewer))

    if len(application.viewers) != 1:
        raise ValueError("Plotly Export only supports a single tab. "
                         "Please close other tabs to export")

    nplot = sum(len(t) for t in application.viewers)
    if nplot == 0:
        raise ValueError("Plotly Export requires at least one plot")

    if nplot > 4:
        raise ValueError("Plotly Export supports at most 4 plots")


DISPATCH = {}



from __future__ import absolute_import, division, print_function

import sys

from qtpy import QtCore

__all__ = ['Worker']


class Worker(QtCore.QThread):

    result = QtCore.Signal(object)
    error = QtCore.Signal(object)

    def __init__(self, func, *args, **kwargs):
        """
        Execute a function call on a different thread.

        Parameters
        ----------
        func : callable
            The function object to call
        args
            Positional arguments to pass to the function
        kwargs
            Keyword arguments to pass to the function
        """
        super(Worker, self).__init__()
        self.func = func
        self.args = args or ()
        self.kwargs = kwargs or {}

    def run(self):
        """
        Invoke the function. Upon successful completion, the result signal
        will be fired with the output of the function If an exception
        occurs, the error signal will be fired with the result form
        sys.exc_info()
        """
        try:
            self.running = True
            result = self.func(*self.args, **self.kwargs)
            self.running = False
            self.result.emit(result)
        except Exception:
            self.running = False
            self.error.emit(sys.exc_info())
